/* Copyright (C) 2004,2005 David Antliff <dave.antliff@paradise.net.nz>
 * Copyright (C) 2006-2009 David Mohr <david@mcbf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 *
 * With a dual monitor setup, where each monitor is its own X screen and
 * is seperated so that there is no automatic transition, this program allows
 * you to turn that on when needed. So you can control when you want to let 
 * the mouse jump over at the edge of the screen to the other one.
 *
 * Right now only two monitors connected horizontally is supported.
 * Also, the mouse jumps over both edges of the screen (a wrap-around). If you
 * really don't want that, email me and I'll see what I can do for you.
 *
 * Please email me any comments, bugs, etc.
 * 
 * Version 0.5 (2009-05-05) - Rename to mouse-wrapscreen.
 *                            Integrated patch by Ng Oon-Ee <ngoonee@gmail.com>, which in turn was 
 *                            based on a program by Sampo Niskanen <sampo.niskanen@removethis.iki.fi>
 *                            to use XWrapPointer instead of the XTest extension,
 *                            which broke during an xorg upgrade.
 *                            Note: 0.4 was skipped.
 *
 * Version 0.3b(2008-08-15) - Bugfix: place cursor on width - 2 of the other screen,
 *                            not this screen.
 *
 * Version 0.3 (2008-08-13) - Bugfix: correctly work with screens of different
 *                            sizes.
 *
 * Version 0.2 (2007-02-27) - Added switches to define screen relation,
 *                            updated code to correctly detect which screen
 *                            the mouse is currently on.
 *
 * Compile with:
 *  gcc -g -Wall mouse-switchscreen.c -o mouse-switchscreen \\
 *      -L /usr/X11R6/lib -lX11 -lXtst -lXext
 * 
 * This program was heavily based on 
 *    Jail, by Sebastien Marion <seb.marion@gmail.com>, 
 *  which in turn was heavily based on 
 *    switchscreen, by David Antliff <dave.antliff@paradise.net.nz>,
 * and mouse-switchscreen is by David Mohr <david@mcbf.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <X11/Xlib.h>

#define PROG_NAME "mouse-wrapscreen"
#define PRINT_PREFIX PROG_NAME ": "
#define VERSION "0.5"
#define DEFAULT_RESISTANCE 15

inline int
getOtherScreen (int screen)
{
  return 1 - screen;
}

static struct option const long_options[] = {
  {"help", 0, NULL, 'h'},
  {"resistance", no_argument, NULL, 'r'},
  {"version", no_argument, NULL, 'V'},
  {"verbose", no_argument, NULL, 'v'},
  {"left", no_argument, NULL, 'l'},
  {"right", no_argument, NULL, 'i'},
  {"both", no_argument, NULL, 'b'},
  {NULL, 0, NULL, 0}
};

void
usage (char **argv)
{
  printf ("Usage: %s\n"
          "    -V, --version       Print version and exit\n"
          "    -r, --resistance=R  Mouse pointer has an edge resistance of R when cross-,\n"
          "                        ing from one screen to the other (%d is the default,\n"
          "                        1 will make the mouse pass over right away)\n"
          "    -v, --verbose       Print some information about what is being done\n"
          "                        Can be specified twice for even more outputd\n"
          "    -h, --help          Print this help text and exit\n"
          "\n"
          " You can define the relation of the screens, if you the cursor to only pass\n"
          " one edge:\n"
          "    -l, --left          Screen 1 is left of screen 0\n"
          "    -i, --right         Screen 1 is rigt of screen 0\n"
          "    -b, --both          Pass cursor on both the left and the right edge\n"
          "                        (default)\n", argv[0], DEFAULT_RESISTANCE);
}

int
main (int argc, char **argv)
{
  Display *display;
  Window window;
  char *displayName;
  int thisScreen;
  int otherScreen = -1;
  char c;
  int resistance = DEFAULT_RESISTANCE;
  int verbose = 0;
  int left = 1;
  int right = 1;

  while ((c = getopt_long (argc, argv, "r:hvlib", long_options, NULL)) != -1) {
    switch (c) {
    case 'h':
      usage (argv);
      return 0;
      break;
    case 'V':
      printf (PROG_NAME " %s, GPL\n", VERSION);
      return 0;
      break;
    case 'r':
      resistance = atoi (optarg);
      if (resistance <= 1) {
        fprintf (stderr, PRINT_PREFIX "Edge resistance must be greater or equal to one\n");
        usage (argv);
        return 1;
      }
      break;
    case 'v':
      verbose++;
      break;
    case 'l':
      left = 1;
      right = 0;
      break;
    case 'i':
      left = 0;
      right = 1;
      break;
    case 'b':
      left = right = 1;
      break;
    default:
      if (c != '?')
        fprintf (stderr, PRINT_PREFIX "Unknown parameter %c\n", c);
      usage (argv);
      return 1;
    }
  }


  displayName = getenv ("DISPLAY");
  if (displayName == NULL) {
    fprintf (stderr, PRINT_PREFIX "DISPLAY is not set\n");
    exit (1);
  }

  if (verbose)
    fprintf (stderr, PRINT_PREFIX "Opening display %s\n", displayName);
  display = XOpenDisplay (displayName);

  if (ScreenCount (display) != 2) {
    printf ("This program only works for two screens at the moment, sorry\n");
    return 2;
  }

  if (verbose) {
    fprintf (stderr, PRINT_PREFIX "Using a resistance of %d\n", resistance);
    if (left && right)
      fprintf (stderr, PRINT_PREFIX "Crossing both screen edges\n");
    else if (left)
      fprintf (stderr, PRINT_PREFIX "Screen 1 is left of screen 0\n");
    else if (right)
      fprintf (stderr, PRINT_PREFIX "Screen 1 is right of screen 0\n");
  }

  thisScreen = DefaultScreen (display);
  //fprintf(stderr,"Started up on screen %d\n",thisScreen);

  otherScreen = getOtherScreen (thisScreen);

  if (otherScreen != thisScreen) {

    int root_x, root_y;
    int win_x, win_y;
    unsigned int mask;
    Window root, child;
    int curscreen, newscreen;
    int num_screens = 2;
    int w[num_screens], h[num_screens];
    int modif = 0;
    int reset = 0;
    Bool r;

    if (thisScreen >= num_screens || otherScreen >= num_screens) {
      fprintf (stderr, PRINT_PREFIX "Invalid screen configuration (%d vs. %d)! Bug the other for support.\n", thisScreen, otherScreen);
      return 1;
    }

    w[thisScreen] = DisplayWidth (display, thisScreen);
    h[thisScreen] = DisplayHeight (display, thisScreen);

    w[otherScreen] = DisplayWidth (display, otherScreen);
    h[otherScreen] = DisplayHeight (display, otherScreen);

    if (verbose) {
      fprintf (stderr, PRINT_PREFIX "Screen %d has size %dx%d\n", thisScreen, w[thisScreen], h[thisScreen]);
      fprintf (stderr, PRINT_PREFIX "Screen %d has size %dx%d\n", otherScreen, w[otherScreen], h[otherScreen]);
    }

    XQueryPointer (display, RootWindow (display, thisScreen), &root, &child, &root_x, &root_y, &win_x, &win_y, &mask);

    while (1) {
      // get current mouse coordinates
      r = XQueryPointer (display,
                         RootWindow (display, thisScreen), &root, &child, &root_x, &root_y, &win_x, &win_y, &mask);
      // r==true if the cursor is in the screen we started up with

      if (r == 1)
        curscreen = thisScreen;
      else
        curscreen = otherScreen;

      newscreen = getOtherScreen (curscreen);

      if (root_x == 0 && ((curscreen == 0 && left) || (curscreen == 1 && right))) {
        reset = 0;
        modif++;
        root_x = w[newscreen] - 2;
        if (verbose > 1)
          fprintf(stderr, PRINT_PREFIX "Leftwards - Count = %d, r=%d\n", modif, r);
      }
      else if (root_x == w[curscreen] - 1 && ((curscreen == 1 && left) || (curscreen == 0 && right))) {
        reset = 0;
        modif++;
        root_x = 1;
        if (verbose > 1)
          fprintf(stderr, PRINT_PREFIX "Rightwards - Count = %d, r=%d\n", modif,r );
      }

      //fprintf(stderr,"display: %d, thisScreen: %d, root: %d, root_x: %d, root_y: %d\n",display,thisScreen,root, root_x,root_y);

      // check if we should wrap the cursor
      if (modif >= resistance) {
        if (verbose)
          fprintf (stderr, PRINT_PREFIX "Switching mouse over to screen %d at [%4d,%4d]\n", newscreen, root_x, root_y);

        window = RootWindow(display, newscreen);

#if 0
        // We don't want to grab the pointer, else the cursor move is not noticed by the wm, and
        // focus follow mouse breaks. Thanks ofourdan.
        r = XGrabPointer(display, window, False, 0, GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
        if (r != GrabSuccess) {
          fprintf (stderr, PRINT_PREFIX "Error grabbing the pointer (%d)", r);
          continue;
        }
#endif

        r = XWarpPointer(display, None, window, 0, 0, 0, 0, root_x, root_y);
        if (r != 1)
          fprintf (stderr, PRINT_PREFIX "Error warping the pointer");
        
#if 0
        XUngrabPointer(display, CurrentTime);
#endif

        modif = 0;
      }

      //sleeps for 1ms                 
      usleep (1000);

      // reset gets triggered when we see an unrelated event
      if (reset)
        modif = 0;

      reset = 1;
    }

  }
  XCloseDisplay (display);

  return (0);
}
